@@ -27,7 +27,7 @@ static char		**pickargv;
  */
 static const char	 *pickenv[] = {
 	"LC_CTYPE",		"en_US.UTF-8",
-	"MALLOC_OPTIONS",	"S",		/* malloc.conf(5) options on OpenBSD */
+	"MALLOC_OPTIONS",	"RS",		/* malloc.conf(5) options on OpenBSD */
 	"TERM",			"xterm",
 	NULL,
 };
@@ -196,7 +196,7 @@ parent(int master, int slave, const char *keys)
 	len = strlen(keys);
 
 	memset(&timeout, 0, sizeof(timeout));
-	timeout.tv_sec = 1;
+	timeout.tv_sec = 2;
 	while (gotsig == 0) {
 		FD_ZERO(&rfd);
 		FD_SET(master, &rfd);
